"""Bond BPUP wrapper."""

import asyncio
import logging
import time
from typing import Any, Callable, Dict, List, Optional, cast

import orjson

BPUP_INIT_PUSH_MESSAGE = b"\n"
BPUP_PORT = 30007
BPUP_ALIVE_TIMEOUT = 70

_LOGGER = logging.getLogger(__name__)


class BPUPSubscriptions:
    """Store BPUP subscriptions."""

    def __init__(self) -> None:
        """Init and store callbacks."""
        self._callbacks: Dict[str, List[Callable]] = {}
        self.last_message_time: float = -BPUP_ALIVE_TIMEOUT

    @property
    def alive(self) -> bool:
        """Return if the subscriptions are considered alive."""
        return (time.monotonic() - self.last_message_time) < BPUP_ALIVE_TIMEOUT

    def connection_lost(self) -> None:
        """Set the last message time to never."""
        self.last_message_time = -BPUP_ALIVE_TIMEOUT

    def subscribe(self, device_id: str, callback: Callable) -> None:
        """Subscribe to BPUP updates."""
        self._callbacks.setdefault(device_id, []).append(callback)

    def unsubscribe(self, device_id: str, callback: Callable) -> None:
        """Unsubscribe from BPUP updates."""
        self._callbacks[device_id].remove(callback)

    def notify(self, json_msg: Dict[str, Any]) -> None:
        """Notify subscribers of an update."""
        self.last_message_time = time.monotonic()

        if json_msg.get("s") != 200:
            return

        topic = json_msg["t"].split("/")
        device_id = topic[1]

        for callback in self._callbacks.get(device_id, []):
            callback(json_msg)


class BPUProtocol(asyncio.Protocol):
    """Implements BPU Protocol."""

    def __init__(self, bpup_subscriptions: BPUPSubscriptions) -> None:
        """Create BPU Protocol."""
        self.loop = asyncio.get_event_loop()
        self.bpup_subscriptions: BPUPSubscriptions = bpup_subscriptions
        self.transport: Optional[asyncio.DatagramTransport] = None
        self.keep_alive: Optional[asyncio.TimerHandle] = None

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        """Connect or reconnect to the device."""
        self.transport = cast(asyncio.DatagramTransport, transport)
        if self.keep_alive:
            self.keep_alive.cancel()
            self.keep_alive = None
        self.send_keep_alive()

    def send_keep_alive(self) -> None:
        """Send a keep alive every 60 seconds per the protocol."""
        if not self.transport or self.transport.is_closing():
            return
        self.transport.sendto(BPUP_INIT_PUSH_MESSAGE)
        self.keep_alive = self.loop.call_later(60, self.send_keep_alive)

    def datagram_received(self, data: bytes, addr: Any) -> None:
        """Process incoming state changes."""
        _LOGGER.debug("%s: BPUP message: %s", addr, data)
        try:
            self.bpup_subscriptions.notify(orjson.loads(data.decode().rstrip("\n")))
        except orjson.JSONDecodeError as ex:
            _LOGGER.warning(
                "%s: Failed to process BPUP message: %s: %s", addr, data, ex
            )

    def error_received(self, exc: Optional[Exception]) -> None:
        """Log errors."""
        assert self.transport is not None
        _LOGGER.error(
            "BPUP error (peer:%s sock:%s): %s",
            self.transport.get_extra_info("peername"),
            self.transport.get_extra_info("sockname"),
            exc,
        )

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """Log connection lost."""
        assert self.transport is not None
        self.bpup_subscriptions.connection_lost()
        if exc:
            _LOGGER.error(
                "BPUP connection lost (peer:%s sock:%s): %s",
                self.transport.get_extra_info("peername"),
                self.transport.get_extra_info("sockname"),
                exc,
            )

    def stop(self) -> None:
        """Stop the client."""
        _LOGGER.debug("BPUP connection stopping: %s", self.transport)
        self.bpup_subscriptions.connection_lost()
        if self.transport:
            self.transport.close()


async def start_bpup(
    host_ip_addr: str, bpup_subscriptions: BPUPSubscriptions
) -> Callable:
    """Create the socket and protocol."""
    loop = asyncio.get_event_loop()

    _, protocol = await loop.create_datagram_endpoint(
        lambda: BPUProtocol(bpup_subscriptions),
        remote_addr=(host_ip_addr, BPUP_PORT),
    )
    bpup_protocol = cast(BPUProtocol, protocol)
    return bpup_protocol.stop



"""Bond Local API wrapper."""
import random
import uuid
from typing import Any, Callable, List, Optional
from xmlrpc.client import Boolean

import orjson
from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientOSError, ServerDisconnectedError

from bond_async.bond_type import BondType
from .action import Action
from .requestor_uuid import RequestorUUID


class Bond:
    """Bond API."""

    def __init__(
            self,
            host: str,
            token: str,
            requestor_uuid: RequestorUUID = RequestorUUID.ANONYMOUS,
            *,
            session: Optional[ClientSession] = None,
            timeout: Optional[ClientTimeout] = None,
    ):
        """Initialize Bond with provided host and token."""
        if not requestor_uuid.is_allowed():
            raise ValueError(f"Requestor UUID {requestor_uuid} is not allowed. Please use a requestor UUID with a "
                             f"number greater than 0xA0.")
        self._requestor_uuid = requestor_uuid
        self._session_uuid = uuid.uuid4().hex[:4]
        self._host = host
        self._api_kwargs = {"headers": {"BOND-Token": token}}
        if timeout:
            self._api_kwargs["timeout"] = timeout
        self._session = session

    async def version(self) -> dict:
        """Return the version of Bond reported by API."""
        return await self.__get("/v2/sys/version")

    async def bond_type(self) -> BondType:
        """Return the BondType based on the serial number reported by API."""
        version = await self.version()
        return BondType.from_serial(version["bondid"])

    async def token(self) -> dict:
        """Return the token after power rest or proof of ownership event."""
        return await self.__get("/v2/token")

    async def bridge(self) -> dict:
        """Return the name and location of the bridge."""
        return await self.__get("/v2/bridge")

    async def set_bluelight_brightness(self, brightness: int) -> None:
        """Set the brightness of the blue light on the bridge. Accepts values from 0 to 255."""
        if brightness < 0 or brightness > 255:
            raise ValueError("Brightness must be between 0 and 255")
        await self.__patch("/v2/bridge", {"bluelight": brightness})

    async def devices(self) -> List[str]:
        """Return the list of available device IDs reported by API."""
        json = await self.__get("/v2/devices")
        return [key for key in json if not key.startswith("_") and isinstance(json[key], dict)]

    async def device(self, device_id: str) -> dict:
        """Return main device metadata reported by API."""
        return await self.__get(f"/v2/devices/{device_id}")

    async def device_properties(self, device_id: str) -> dict:
        """Return device properties reported by API."""
        return await self.__get(f"/v2/devices/{device_id}/properties")

    async def device_state(self, device_id: str) -> dict:
        """Return current device state reported by API."""
        return await self.__get(f"/v2/devices/{device_id}/state")

    async def device_skeds(self, device_id: str) -> dict:
        """Return current device schedules reported by API."""
        return await self.__get(f"/v2/devices/{device_id}/skeds")

    async def action(self, device_id: str, action: Action) -> None:
        """Execute given action for a given device."""
        if action.name == Action.SET_STATE_BELIEF:
            path = f"/v2/devices/{device_id}/state"

            async def patch(session: ClientSession) -> None:
                self._api_kwargs["headers"]["BOND-UUID"] = self.__create_message_id()
                async with session.patch(
                        f"http://{self._host}{path}",
                        **self._api_kwargs,
                        json=action.argument,
                ) as response:
                    response.raise_for_status()

            await self.__call(patch)
        else:
            path = f"/v2/devices/{device_id}/actions/{action.name}"

            async def put(session: ClientSession) -> None:
                self._api_kwargs["headers"]["BOND-UUID"] = self.__create_message_id()
                async with session.put(
                        f"http://{self._host}{path}",
                        **self._api_kwargs,
                        json=action.argument,
                ) as response:
                    response.raise_for_status()

            await self.__call(put)

    async def supports_groups(self) -> Boolean:
        """Return 'True' if the Bond supports the Groups feature."""
        json = await self.__get("/v2/")
        return "groups" in json

    async def groups(self) -> List[str]:
        """Return the list of available group IDs reported by API."""
        json = await self.__get("/v2/groups")
        return [key for key in json if not key.startswith("_") and type(json[key]) is dict]

    async def group(self, group_id: str) -> dict:
        """Return main group metadata reported by API."""
        return await self.__get(f"/v2/groups/{group_id}")

    async def group_properties(self, group_id: str) -> dict:
        """Return group properties reported by API."""
        return await self.__get(f"/v2/groups/{group_id}/properties")

    async def group_state(self, group_id: str) -> dict:
        """Return current group state reported by API."""
        return await self.__get(f"/v2/groups/{group_id}/state")

    async def group_skeds(self, group_id: str) -> dict:
        """Return current group schedules reported by API."""
        return await self.__get(f"/v2/groups/{group_id}/skeds")

    async def group_action(self, group_id: str, action: Action) -> None:
        """Execute given action for a given group."""
        if action.name == Action.SET_STATE_BELIEF:
            path = f"/v2/groups/{group_id}/state"

            async def patch(session: ClientSession) -> None:
                self._api_kwargs["headers"]["BOND-UUID"] = self.__create_message_id()
                async with session.patch(
                        f"http://{self._host}{path}",
                        **self._api_kwargs,
                        json=action.argument,
                ) as response:
                    response.raise_for_status()

            await self.__call(patch)
        else:
            path = f"/v2/groups/{group_id}/actions/{action.name}"

            async def put(session: ClientSession) -> None:
                self._api_kwargs["headers"]["BOND-UUID"] = self.__create_message_id()
                async with session.put(
                        f"http://{self._host}{path}",
                        **self._api_kwargs,
                        json=action.argument,
                ) as response:
                    response.raise_for_status()

            await self.__call(put)

    async def __get(self, path) -> dict:
        async def get(session: ClientSession) -> dict:
            self._api_kwargs["headers"]["BOND-UUID"] = self.__create_message_id()
            async with session.get(
                    f"http://{self._host}{path}", **self._api_kwargs
            ) as response:
                response.raise_for_status()
                return await response.json(loads=orjson.loads)

        return await self.__call(get)

    async def __patch(self, path, json) -> None:
        async def patch(session: ClientSession) -> None:
            self._api_kwargs["headers"]["BOND-UUID"] = self.__create_message_id()
            async with session.patch(
                f"http://{self._host}{path}",
                **self._api_kwargs,
                json=json,
            ) as response:
                response.raise_for_status()

        await self.__call(patch)

    async def __put(self, path, json) -> None:
        async def put(session: ClientSession) -> None:
            self._api_kwargs["headers"]["BOND-UUID"] = self.__create_message_id()
            async with session.put(
                f"http://{self._host}{path}",
                **self._api_kwargs,
                json=json,
            ) as response:
                response.raise_for_status()

        await self.__call(put)

    async def __call(self, handler: Callable[[ClientSession], Any]):
        if not self._session:
            async with ClientSession() as request_session:
                return await handler(request_session)
        else:
            try:
                return await handler(self._session)
            except (ClientOSError, ServerDisconnectedError):
                # bond has a short connection close time
                # so we need to retry if we idled for a bit
                return await handler(self._session)

    def __create_message_id(self) -> str:
        """Create a unique hex message ID.
        The first 2 characters is the requestor_uuid (in hex),
        the next 4 characters are always the same for a session,
        and the last 10 characters are random."""
        return (
            f"{self._requestor_uuid.hex_value()}"
            f"{self._session_uuid}"
            f"{random.randint(0, 0xFFFFFFFF):010x}"
        ).lower()



"""bond-async unit tests."""

from datetime import datetime
import asyncio
import time


def mock_time_changed(loop: asyncio.AbstractEventLoop, datetime_: datetime) -> None:
    """Call events in the future."""
    for task in list(loop._scheduled):
        if not isinstance(task, asyncio.TimerHandle):
            continue
        if task.cancelled():
            continue

        mock_seconds_into_future = datetime_.timestamp() - time.time()
        future_seconds = task.when() - loop.time()

        if mock_seconds_into_future >= future_seconds:
            task._run()
            task.cancel()



import asyncio
import logging

from bond_async import BPUPSubscriptions, start_bpup


async def main(ip_address):
    """Example of library usage."""

    sub = BPUPSubscriptions()
    stop_bpup = await start_bpup(ip_address, sub)

    for i in range(500):
        print("BPUP is alive:", sub.alive)
        await asyncio.sleep(1)

    stop_bpup()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)

    print("Enter the device ip:")
    ip_address = input().strip()
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main(ip_address))



from enum import Enum


class RequestorUUID(Enum):
    # These are the existing requestor IDs that are already allocated
    TEST = 0x00
    # Only allow requestor IDs greater than 0xA0
    ANONYMOUS = 0xFF
    HOME_ASSISTANT = 0xA0

    # method that returns if a given hex requestor is allowed
    def is_allowed(self) -> bool:
        return self.value >= 0xA0

    def hex_value(self) -> str:
        return f"{self.value:02x}"



"""Unit tests for Action."""

from bond_async import Action


def test_action_eq():
    """Tests that different Action instances compare correctly."""
    assert Action("name-1", argument="arg-1") == Action("name-1", argument="arg-1")
    assert Action("name-1", argument="arg-1") != Action("name-other", argument="arg-1")
    assert Action("name-1", argument="arg-1") != Action("name-1", argument="arg-other")
    assert Action("name-1", argument="arg-1") != Action(
        "name-other", argument="arg-other"
    )



"""Bond type enumeration."""
import re
from enum import Enum

regexes = {
    "bridge_snowbird": r"^[A-C]\w*$",
    "bridge_zermatt": r"^Z(Z|X)\w*$",
    "bridge_pro": r"^ZP\w*$",
    "sbb_lights": r"^T\w*$",
    "sbb_ceiling_fan": r"^K\w*$",
    "sbb_plug": r"^P\w*$",
}


class BondType(Enum):
    """Bond type enumeration."""

    BRIDGE_SNOWBIRD = "bridge_snowbird"
    BRIDGE_ZERMATT = "bridge_zermatt"
    BRIDGE_PRO = "bridge_pro"
    SBB_LIGHTS = "sbb_lights"
    SBB_CEILING_FAN = "sbb_ceiling_fan"
    SBB_PLUG = "sbb_plug"

    def is_sbb(self) -> bool:
        """Checks if BondType is a Smart by Bond product."""
        return self.value.startswith("sbb_")

    def is_bridge(self) -> bool:
        """Checks if BondType is a Bond Bridge/Bond Bridge Pro."""
        return self.value.startswith("bridge_")

    @classmethod
    def from_serial(cls, serial: str):
        """Returns a BondType for a serial number"""
        for (bond_type, regex) in regexes.items():
            if re.search(regex, serial):
                return cls(bond_type)
        return None

    @staticmethod
    def is_sbb_from_serial(serial: str) -> bool:
        """Checks if specified Bond serial number is a Smart by Bond product."""
        bond_type = BondType.from_serial(serial)
        if bond_type:
            return bond_type.is_sbb()
        return False

    @staticmethod
    def is_bridge_from_serial(serial: str) -> bool:
        """Checks if specified Bond serial number is a Bond Bridge."""
        bond_type = BondType.from_serial(serial)
        if bond_type:
            return bond_type.is_bridge()
        return False



"""Unit tests for Bond API wrapper."""

import pytest
from aiohttp import ClientSession, ClientTimeout
from aioresponses import CallbackResult, aioresponses

from bond_async import Action, Bond, Direction, RequestorUUID
from bond_async.bond_type import BondType


@pytest.fixture(name="bond")
def bond_fixture():
    """Creates Bond fixture."""
    return Bond("test-host", "test-token")


@pytest.mark.asyncio
async def test_optional_overrides():
    """Tests using external session."""
    async with ClientSession() as session:
        timeout: ClientTimeout = ClientTimeout(total=1)
        bond: Bond = Bond("test-host", "test-token", session=session, timeout=timeout)
        with aioresponses() as response:
            response.get("http://test-host/v2/sys/version", payload={"some": "version"})
            actual = await bond.version()
            assert actual == {"some": "version"}


@pytest.mark.asyncio
async def test_illegal_requestor_uuid():
    """Tests using illegal requestor UUID."""
    with pytest.raises(ValueError):
        Bond("test-host", "test-token", requestor_uuid=RequestorUUID.TEST)


@pytest.mark.asyncio
async def test_legal_requestor_uuid():
    """Tests using legal requestor UUID."""
    bond = Bond("test-host", "test-token", requestor_uuid=RequestorUUID.ANONYMOUS)
    assert bond._requestor_uuid == RequestorUUID.ANONYMOUS


@pytest.mark.asyncio
async def test_create_message_id():
    bond = Bond("test-host", "test-token", requestor_uuid=RequestorUUID.ANONYMOUS)
    message_id = bond._Bond__create_message_id()
    session_id = bond._session_uuid
    assert message_id[:2] == "ff"
    assert message_id[2:6] == session_id
    assert len(message_id) == 16

    # a new call should generate a new message id, but the requestor uuid and session id should be the same
    new_message_id = bond._Bond__create_message_id()
    assert new_message_id[:2] == "ff"
    assert new_message_id[2:6] == session_id
    assert len(new_message_id) == 16
    assert new_message_id != message_id


@pytest.mark.asyncio
async def test_msg_id_in_request_header(bond: Bond):
    """Tests that the message id is in the request header."""
    with aioresponses() as response:
        response.get(
            "http://test-host/v2/sys/version",
            payload={"some": "version"},
        )
        await bond.version()
        headers = next(iter(response.requests.values()))[0].kwargs["headers"]
        assert "BOND-UUID" in headers
        msg_id = headers["BOND-UUID"]
        assert msg_id[:2] == "ff"
        assert msg_id[2:6] == bond._session_uuid
        assert len(msg_id) == 16


@pytest.mark.asyncio
async def test_version(bond: Bond):
    """Tests version API."""
    with aioresponses() as response:
        response.get("http://test-host/v2/sys/version", payload={"some": "version"})
        actual = await bond.version()
        assert actual == {"some": "version"}


@pytest.mark.asyncio
async def test_bond_type(bond: Bond):
    """Tests version API."""
    with aioresponses() as response:
        response.get("http://test-host/v2/sys/version", payload={"bondid": "KSMJWCE12345"})
        actual = await bond.bond_type()
        assert actual == BondType.SBB_CEILING_FAN


@pytest.mark.asyncio
async def test_bridge(bond: Bond):
    """Tests bridge API."""
    with aioresponses() as response:
        response.get(
            "http://test-host/v2/bridge",
            payload={"name": "name", "location": "location"},
        )
        actual = await bond.bridge()
        assert actual == {"name": "name", "location": "location"}


@pytest.mark.asyncio
async def test_token(bond: Bond):
    """Tests token API."""
    with aioresponses() as response:
        response.get(
            "http://test-host/v2/token",
            payload={"locked": 0, "token": "8f514567acaf9869"},
        )
        actual = await bond.token()
        assert actual == {"locked": 0, "token": "8f514567acaf9869"}


@pytest.mark.asyncio
async def test_devices(bond: Bond):
    """Tests API to get a list of device IDs."""
    with aioresponses() as response:
        response.get(
            "http://test-host/v2/devices",
            payload={
                "_": "some-hash",
                "__": "some-other-hash",
                "device-1": {"_": "some-hash"},
                "device-2": {"_": "some-hash"},
            },
        )
        actual = await bond.devices()
        assert actual == ["device-1", "device-2"]


@pytest.mark.asyncio
async def test_device(bond: Bond):
    """Tests API to get device details."""
    with aioresponses() as response:
        response.get("http://test-host/v2/devices/device-1", payload={"some": "device details"})
        actual = await bond.device("device-1")
        assert actual == {"some": "device details"}


@pytest.mark.asyncio
async def test_device_properties(bond: Bond):
    """Tests API to get device properties."""
    with aioresponses() as response:
        response.get(
            "http://test-host/v2/devices/device-1/properties",
            payload={"some": "device properties"},
        )
        actual = await bond.device_properties("device-1")
        assert actual == {"some": "device properties"}


@pytest.mark.asyncio
async def test_device_state(bond: Bond):
    """Tests API to get device state."""
    with aioresponses() as response:
        response.get(
            "http://test-host/v2/devices/device-1/state",
            payload={"some": "device state"},
        )
        actual = await bond.device_state("device-1")
        assert actual == {"some": "device state"}


@pytest.mark.asyncio
async def test_turn_on(bond: Bond):
    """Tests turn_on action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/TurnOn",
            callback=callback,
        )
        await bond.action("test-device-id", Action.turn_on())


@pytest.mark.asyncio
async def test_turn_off(bond: Bond):
    """Tests turn_off action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/TurnOff",
            callback=callback,
        )
        await bond.action("test-device-id", Action.turn_off())


@pytest.mark.asyncio
async def test_open(bond: Bond):
    """Tests open action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put("http://test-host/v2/devices/test-device-id/actions/Open", callback=callback)
        await bond.action("test-device-id", Action.open())


@pytest.mark.asyncio
async def test_tilt_open(bond: Bond):
    """Tests tilt open action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/TiltOpen",
            callback=callback,
        )
        await bond.action("test-device-id", Action.tilt_open())


@pytest.mark.asyncio
async def test_close(bond: Bond):
    """Tests close action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/Close",
            callback=callback,
        )
        await bond.action("test-device-id", Action.close())


@pytest.mark.asyncio
async def test_tilt_close(bond: Bond):
    """Tests tilt close action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/TiltClose",
            callback=callback,
        )
        await bond.action("test-device-id", Action.tilt_close())


@pytest.mark.asyncio
async def test_hold(bond: Bond):
    """Tests hold action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put("http://test-host/v2/devices/test-device-id/actions/Hold", callback=callback)
        await bond.action("test-device-id", Action.hold())


@pytest.mark.asyncio
async def test_set_speed(bond: Bond):
    """Tests set_speed action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 2}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/SetSpeed",
            callback=callback,
        )
        await bond.action("test-device-id", Action.set_speed(2))


@pytest.mark.asyncio
async def test_set_speed_belief(bond: Bond):
    """Tests set_speed_belief action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"speed": 2}
            return CallbackResult()

        response.patch("http://test-host/v2/devices/test-device-id/state", callback=callback)
        await bond.action("test-device-id", Action.set_speed_belief(2))


@pytest.mark.asyncio
async def test_set_brightness_belief(bond: Bond):
    """Tests set_brightness_belief action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"brightness": 2}
            return CallbackResult()

        response.patch("http://test-host/v2/devices/test-device-id/state", callback=callback)
        await bond.action("test-device-id", Action.set_brightness_belief(2))


@pytest.mark.asyncio
async def test_set_power_state_belief(bond: Bond):
    """Tests set_power_state_belief action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"power": 1}
            return CallbackResult()

        response.patch("http://test-host/v2/devices/test-device-id/state", callback=callback)
        await bond.action("test-device-id", Action.set_power_state_belief(True))


@pytest.mark.asyncio
async def test_set_light_state_belief(bond: Bond):
    """Tests set_light_state_belief action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"light": 0}
            return CallbackResult()

        response.patch("http://test-host/v2/devices/test-device-id/state", callback=callback)
        await bond.action("test-device-id", Action.set_light_state_belief(False))


@pytest.mark.asyncio
async def test_turn_light_on(bond: Bond):
    """Tests turn_light_on action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/TurnLightOn",
            callback=callback,
        )
        await bond.action("test-device-id", Action.turn_light_on())


@pytest.mark.asyncio
async def test_turn_light_off(bond: Bond):
    """Tests turn_light_off action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/TurnLightOff",
            callback=callback,
        )
        await bond.action("test-device-id", Action.turn_light_off())


@pytest.mark.asyncio
async def test_set_brightness(bond: Bond):
    """Tests set_brightness action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 50}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/SetBrightness",
            callback=callback,
        )
        await bond.action("test-device-id", Action.set_brightness(50))


@pytest.mark.asyncio
async def test_set_color_temperature(bond: Bond):
    """Tests set_color_temperature action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 3000}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/SetColorTemp",
            callback=callback,
        )
        await bond.action("test-device-id", Action.set_color_temperature(3000))


@pytest.mark.asyncio
async def test_increase_color_temperature(bond: Bond):
    """Tests increase_color_temperature action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 100}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/IncreaseColorTemp",
            callback=callback,
        )
        await bond.action("test-device-id", Action.increase_color_temperature(100))


@pytest.mark.asyncio
async def test_decrease_color_temperature(bond: Bond):
    """Tests decrease_color_temperature action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 100}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/DecreaseColorTemp",
            callback=callback,
        )
        await bond.action("test-device-id", Action.decrease_color_temperature(100))


@pytest.mark.asyncio
async def test_set_direction_forward(bond: Bond):
    """Tests set_direction action delegates to API with correct value for forward."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 1}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/SetDirection",
            callback=callback,
        )
        await bond.action("test-device-id", Action.set_direction(Direction.FORWARD))


@pytest.mark.asyncio
async def test_set_direction_reverse(bond: Bond):
    """Tests set_direction action delegates to API with correct value for reverse."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": -1}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/SetDirection",
            callback=callback,
        )
        await bond.action("test-device-id", Action.set_direction(Direction.REVERSE))


@pytest.mark.asyncio
async def test_set_flame(bond: Bond):
    """Tests set_flame action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 50}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/SetFlame",
            callback=callback,
        )
        await bond.action("test-device-id", Action.set_flame(50))


@pytest.mark.asyncio
async def test_set_position(bond: Bond):
    """Tests set_position action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 50}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/SetPosition",
            callback=callback,
        )
        await bond.action("test-device-id", Action.set_position(50))


@pytest.mark.asyncio
async def test_increase_position(bond: Bond):
    """Tests increase_position action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 50}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/IncreasePosition",
            callback=callback,
        )
        await bond.action("test-device-id", Action.increase_position(50))


@pytest.mark.asyncio
async def test_decrease_position(bond: Bond):
    """Tests decrease_position action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 50}
            return CallbackResult()

        response.put(
            "http://test-host/v2/devices/test-device-id/actions/DecreasePosition",
            callback=callback,
        )
        await bond.action("test-device-id", Action.decrease_position(50))


@pytest.mark.asyncio
async def test_set_bluelight_brightness(bond: Bond):
    """Tests set_bluelight_brightness action delegates to API."""
    with aioresponses() as response:
        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"bluelight": 50}
            return CallbackResult()

        response.patch(
            "http://test-host/v2/bridge",
            callback=callback,
        )
        await bond.set_bluelight_brightness(50)



"""Bond Device type enumeration."""


class DeviceType:
    """Bond Device type enumeration."""

    CEILING_FAN = "CF"
    MOTORIZED_SHADES = "MS"
    FIREPLACE = "FP"
    AIR_CONDITIONER = "AC"
    GARAGE_DOOR = "GD"
    BIDET = "BD"
    LIGHT = "LT"
    GENERIC_DEVICE = "GX"

    @staticmethod
    def is_fan(device_type: str) -> bool:
        """Checks if specified device type is a fan."""
        return device_type == DeviceType.CEILING_FAN

    @staticmethod
    def is_shades(device_type: str) -> bool:
        """Checks if specified device type is shades."""
        return device_type == DeviceType.MOTORIZED_SHADES

    @staticmethod
    def is_fireplace(device_type: str) -> bool:
        """Checks if specified device type is fireplace."""
        return device_type == DeviceType.FIREPLACE

    @staticmethod
    def is_air_conditioner(device_type: str) -> bool:
        """Checks if specified device type is air conditioner."""
        return device_type == DeviceType.AIR_CONDITIONER

    @staticmethod
    def is_garage_door(device_type: str) -> bool:
        """Checks if specified device type is garage door."""
        return device_type == DeviceType.GARAGE_DOOR

    @staticmethod
    def is_bidet(device_type: str) -> bool:
        """Checks if specified device type is bidet."""
        return device_type == DeviceType.BIDET

    @staticmethod
    def is_light(device_type: str) -> bool:
        """Checks if specified device type is light."""
        return device_type == DeviceType.LIGHT

    @staticmethod
    def is_generic(device_type: str) -> bool:
        """Checks if specified device type is generic."""
        return device_type == DeviceType.GENERIC_DEVICE



"""Unit tests for DeviceType."""

from bond_async.bond_type import BondType


def test_regexes():
    """Tests getting a BondType enum from Bond serial number"""
    assert BondType.BRIDGE_ZERMATT == BondType.from_serial("ZXBL12345")
    assert BondType.BRIDGE_ZERMATT == BondType.from_serial("ZZBL29913")

    assert BondType.BRIDGE_PRO == BondType.from_serial("ZPEA77140")
    assert BondType.BRIDGE_PRO == BondType.from_serial("ZPDK64712")

    assert BondType.BRIDGE_SNOWBIRD == BondType.from_serial("AJ12707")
    assert BondType.BRIDGE_SNOWBIRD == BondType.from_serial("BA12707")

    assert BondType.SBB_LIGHTS == BondType.from_serial("TADDAFA98593")
    assert BondType.SBB_LIGHTS == BondType.from_serial("TWCTAXX56966")

    assert BondType.SBB_CEILING_FAN == BondType.from_serial("KMBI10045")
    assert BondType.SBB_CEILING_FAN == BondType.from_serial("KXXX10044")
    assert BondType.SBB_CEILING_FAN == BondType.from_serial("KSMJCD54321")

    assert BondType.SBB_PLUG == BondType.from_serial("PPCTAXX88803")

    assert None == BondType.from_serial("garbage")
    assert None == BondType.from_serial("NOTASERIAL")


def test_is_sbb():
    """Tests if a Bond serial number is a SBB"""
    assert False == BondType.is_sbb_from_serial("ZXBL12345")
    assert False == BondType.is_sbb_from_serial("ZZBL29913")

    assert False == BondType.is_sbb_from_serial("ZPEA77140")
    assert False == BondType.is_sbb_from_serial("ZPDK64712")

    assert False == BondType.is_sbb_from_serial("AJ12707")
    assert False == BondType.is_sbb_from_serial("BA12707")

    assert True == BondType.is_sbb_from_serial("TADDAFA98593")
    assert True == BondType.is_sbb_from_serial("TWCTAXX56966")

    assert True == BondType.is_sbb_from_serial("KMBI10045")
    assert True == BondType.is_sbb_from_serial("KXXX10044")
    assert True == BondType.is_sbb_from_serial("KSMJCD54321")

    assert True == BondType.is_sbb_from_serial("PPCTAXX88803")

    assert False == BondType.is_sbb_from_serial("garbage")
    assert False == BondType.is_sbb_from_serial("NOTASERIAL")


def test_is_bridge():
    """Tests if a Bond serial number is a SBB"""
    assert True == BondType.is_bridge_from_serial("ZXBL12345")
    assert True == BondType.is_bridge_from_serial("ZZBL29913")

    assert True == BondType.is_bridge_from_serial("ZPEA77140")
    assert True == BondType.is_bridge_from_serial("ZPDK64712")

    assert True == BondType.is_bridge_from_serial("AJ12707")
    assert True == BondType.is_bridge_from_serial("BA12707")

    assert False == BondType.is_bridge_from_serial("TADDAFA98593")
    assert False == BondType.is_bridge_from_serial("TWCTAXX56966")

    assert False == BondType.is_bridge_from_serial("KMBI10045")
    assert False == BondType.is_bridge_from_serial("KXXX10044")
    assert False == BondType.is_bridge_from_serial("KSMJCD54321")

    assert False == BondType.is_bridge_from_serial("PPCTAXX88803")

    assert False == BondType.is_bridge_from_serial("garbage")
    assert False == BondType.is_bridge_from_serial("NOTASERIAL")



"""Typed helpers for executable actions for Bond Local API."""

from enum import IntEnum
from typing import Any


class Direction(IntEnum):
    """Direction enumeration for supported fan directions."""

    FORWARD = 1
    REVERSE = -1


class Action:
    """Namespace for Bond API action helpers and constants."""

    # General actions
    STOP = "Stop"
    TURN_ON = "TurnOn"
    TURN_OFF = "TurnOff"
    TOGGLE_POWER = "TogglePower"
    SET_TIMER = "SetTimer"
    SWITCH_MODE = "SwitchMode"
    SET_STATE_BELIEF = "state"
    HOME_SHIELD_ON = "HomeShieldOn"
    HOME_SHIELD_OFF = "HomeShieldOff"
    HOME_SHIELD_TOGGLE = "HomeShieldToggle"

    # Lights
    TURN_LIGHT_ON = "TurnLightOn"
    TURN_LIGHT_OFF = "TurnLightOff"
    TOGGLE_LIGHT = "ToggleLight"
    SET_BRIGHTNESS = "SetBrightness"
    INCREASE_BRIGHTNESS = "IncreaseBrightness"
    DECREASE_BRIGHTNESS = "DecreaseBrightness"
    START_DIMMER = "StartDimmer"
    TURN_UP_LIGHT_ON = "TurnUpLightOn"
    TURN_DOWN_LIGHT_ON = "TurnDownLightOn"
    TURN_UP_LIGHT_OFF = "TurnUpLightOff"
    TURN_DOWN_LIGHT_OFF = "TurnDownLightOff"
    TOGGLE_UP_LIGHT = "ToggleUpLight"
    TOGGLE_DOWN_LIGHT = "ToggleDownLight"
    START_UP_LIGHT_DIMMER = "StartUpLightDimmer"
    START_DOWN_LIGHT_DIMMER = "StartDownLightDimmer"
    START_INCREASING_BRIGHTNESS = "StartIncreasingBrightness"
    START_DECREASING_BRIGHTNESS = "StartDecreasingBrightness"
    SET_UP_LIGHT_BRIGHTNESS = "SetUpLightBrightness"
    SET_DOWN_LIGHT_BRIGHTNESS = "SetDownLightBrightness"
    INCREASE_UP_LIGHT_BRIGHTNESS = "IncreaseUpLightBrightness"
    DECREASE_UP_LIGHT_BRIGHTNESS = "DecreaseUpLightBrightness"
    INCREASE_DOWN_LIGHT_BRIGHTNESS = "IncreaseDownLightBrightness"
    DECREASE_DOWN_LIGHT_BRIGHTNESS = "DecreaseDownLightBrightness"
    CYCLE_UP_LIGHT_BRIGHTNESS = "CycleUpLightBrightness"
    CYCLE_DOWN_LIGHT_BRIGHTNESS = "CycleDownLightBrightness"
    CYCLE_BRIGHTNESS = "CycleBrightness"
    CYCLE_COLOR_TEMP = "CycleColorTemp"
    SET_COLOR_TEMP = "SetColorTemp"
    CYCLE_COLOR_TEMP_PRESET = "CycleColorTempPreset"
    DECREASE_COLOR_TEMP = "DecreaseColorTemp"
    INCREASE_COLOR_TEMP = "IncreaseColorTemp"

    # Fans
    BREEZE_OFF = "BreezeOff"
    BREEZE_ON = "BreezeOn"
    DECREASE_SPEED = "DecreaseSpeed"
    DIM_MODE = "DimMode"
    INCREASE_SPEED = "IncreaseSpeed"
    OEM_HOME_AWAY = "OEMHomeAway"
    OEM_RANDOM = "OEMRandom"
    OEM_RANDOM_TOGGLE = "OEMRandomToggle"
    OEM_TIMER = "OEMTimer"
    OEM_WALK_AWAY = "OEMWalkAway"
    SET_BREEZE = "SetBreeze"
    SET_DIRECTION = "SetDirection"
    SET_SPEED = "SetSpeed"
    TOGGLE_DIRECTION = "ToggleDirection"
    TOGGLE_LIGHT_TEMP = "ToggleLightTemp"

    # Fireplaces
    INCREASE_TEMPERATURE = "IncreaseTemperature"
    DECREASE_TEMPERATURE = "DecreaseTemperature"
    SET_FP_FAN = "SetFpFan"
    TURN_FP_FAN_ON = "TurnFpFanOn"
    TURN_FP_FAN_OFF = "TurnFpFanOff"
    INCREASE_FLAME = "IncreaseFlame"
    DECREASE_FLAME = "DecreaseFlame"
    SET_FLAME = "SetFlame"

    # Motorized Shades
    CLOSE = "Close"
    CLOSE_NEXT = "CloseNext"
    DECREASE_POSITION = "DecreasePosition"
    ESC = "Esc"
    HOLD = "Hold"
    INCREASE_POSITION = "IncreasePosition"
    LONG_CLOSE = "LongClose"
    LONG_HOLD = "LongHold"
    LONG_OPEN = "LongOpen"
    LONG_PAIR = "LongPair"
    OPEN = "Open"
    OPEN_NEXT = "OpenNext"
    PAIR = "Pair"
    PRESET = "Preset"
    SAVE_LIMIT_BOTTOM_OEM = "SaveLimitBottomOEM"
    SAVE_LIMIT_TOP_OEM = "SaveLimitTopOEM"
    SET_LIMIT_BOTTOM_OEM = "SetLimitBottomOEM"
    SET_LIMIT_TOP_OEM = "SetLimitTopOEM"
    SET_POSITION = "SetPosition"
    TILT_CLOSE = "TiltClose"
    TILT_LEFT = "TiltLeft"
    TILT_LEFT_LONG = "TiltLeftLong"
    TILT_OPEN = "TiltOpen"
    TILT_RIGHT = "TiltRight"
    TILT_RIGHT_LONG = "TiltRightLong"
    TILT_STOP = "TiltStop"
    TOGGLE_OPEN = "ToggleOpen"

    def __init__(self, name: str, argument: Any = None):
        self._name = name
        if self._name == Action.SET_STATE_BELIEF:
            self._argument = {} if argument is None else argument
        else:
            self._argument = {} if argument is None else {"argument": argument}

    def __eq__(self, other: "Action"):
        return self.name == other.name and self.argument == other.argument

    @staticmethod
    def turn_on() -> "Action":
        """Turn device on (usually power)."""
        return Action(Action.TURN_ON)

    @staticmethod
    def turn_off() -> "Action":
        """Turn device off (usually power)."""
        return Action(Action.TURN_OFF)

    @staticmethod
    def open() -> "Action":
        """Open cover."""
        return Action(Action.OPEN)

    @staticmethod
    def close() -> "Action":
        """Close cover."""
        return Action(Action.CLOSE)

    @staticmethod
    def tilt_open() -> "Action":
        """Tilt open cover."""
        return Action(Action.TILT_OPEN)

    @staticmethod
    def tilt_close() -> "Action":
        """Tilt close cover."""
        return Action(Action.TILT_CLOSE)

    @staticmethod
    def hold() -> "Action":
        """Hold cover."""
        return Action(Action.HOLD)

    @staticmethod
    def set_speed(speed: int) -> "Action":
        """Sets fan rotation to a provided speed."""
        return Action(Action.SET_SPEED, speed)

    @staticmethod
    def set_speed_belief(speed: int) -> "Action":
        """Sets fan rotation to a provided speed."""
        return Action(Action.SET_STATE_BELIEF, {"speed": speed})

    @staticmethod
    def set_direction(direction: Direction) -> "Action":
        """Sets fan rotation direction."""
        return Action(Action.SET_DIRECTION, direction.value)

    @staticmethod
    def turn_light_on() -> "Action":
        """Turns on the fan light."""
        return Action(Action.TURN_LIGHT_ON)

    @staticmethod
    def turn_light_off() -> "Action":
        """Turns off the fan light."""
        return Action(Action.TURN_LIGHT_OFF)

    @staticmethod
    def set_light_state_belief(state: bool) -> "Action":
        """Sets light state belief, true or false."""
        return Action(Action.SET_STATE_BELIEF, {"light": int(state)})

    @staticmethod
    def set_power_state_belief(state: bool) -> "Action":
        """Sets light state belief, true or false."""
        return Action(Action.SET_STATE_BELIEF, {"power": int(state)})

    @staticmethod
    def set_brightness(brightness: int) -> "Action":
        """Sets brightness of the light as percentage value, 1-100."""
        return Action(Action.SET_BRIGHTNESS, brightness)

    @staticmethod
    def set_color_temperature(temperature: int) -> "Action":
        """Sets color temperature of the light in Kelin. Resolution: 100 K."""
        return Action(Action.SET_COLOR_TEMP, temperature)

    @staticmethod
    def increase_color_temperature(temperature: int) -> "Action":
        """Increases color temperature a specified number of degrees K. Implicitly turn Light on."""
        return Action(Action.INCREASE_COLOR_TEMP, temperature)

    @staticmethod
    def decrease_color_temperature(temperature: int) -> "Action":
        """Decreases color temperature a specified number of degrees K. Implicitly turn Light on."""
        return Action(Action.DECREASE_COLOR_TEMP, temperature)

    @staticmethod
    def set_brightness_belief(brightness: int) -> "Action":
        """Sets brightness belief of the light as percentage value, 1-100."""
        return Action(Action.SET_STATE_BELIEF, {"brightness": brightness})

    @staticmethod
    def set_flame(flame: int) -> "Action":
        """Sets the flame to given intensity in percent."""
        return Action(Action.SET_FLAME, flame)

    @staticmethod
    def set_position(position: int) -> "Action":
        """Sets shade position percentage from 0 (open) to 100 (closed)."""
        return Action(Action.SET_POSITION, position)

    @staticmethod
    def increase_position(position: int) -> "Action":
        """Closes the device by the specified percentage of the full range."""
        return Action(Action.INCREASE_POSITION, position)

    @staticmethod
    def decrease_position(position: int) -> "Action":
        """Opens the device by the specified percentage of the full range."""
        return Action(Action.DECREASE_POSITION, position)

    @property
    def name(self) -> str:
        """Return name of this action."""
        return self._name

    @property
    def argument(self) -> dict:
        """Return optional argument for this action."""
        return self._argument



"""Unit tests for Bond BPUP."""

from unittest.mock import call, MagicMock, patch
from typing import Optional
import asyncio
import pytest
import datetime as dt

from . import mock_time_changed
from bond_async.bpup import BPUPSubscriptions, BPUProtocol, start_bpup

MOCK_ADDR = ("127.0.0.1", 1)


def mock_protocol_connection_lost(
    protocol: asyncio.BaseProtocol, exc: Optional[Exception]
) -> None:
    """Mock an asyncio.Protocol connection lost callback."""
    protocol.connection_lost(exc)
    protocol.transport.is_closing = MagicMock(return_value=True)


@pytest.fixture(name="transport")
def transport_fixture():
    """Creates transport fixture."""
    transport = MagicMock(auto_spec=asyncio.DatagramTransport)
    transport.is_closing = MagicMock(return_value=False)

    def _mock_close(*_):
        transport.is_closing = MagicMock(return_value=True)

    transport.close = _mock_close
    return transport


@pytest.mark.asyncio
async def test_protocol_keep_alive_close(transport):
    bpup_subscriptions = BPUPSubscriptions()
    loop = asyncio.get_event_loop()
    bpup_protocol = BPUProtocol(bpup_subscriptions)

    bpup_protocol.connection_made(transport)
    assert transport.sendto.mock_calls == [call(b"\n")]
    transport.sendto.reset_mock()

    mock_time_changed(loop, dt.datetime.now(dt.timezone.utc) + dt.timedelta(seconds=60))
    assert transport.sendto.mock_calls == [call(b"\n")]
    transport.sendto.reset_mock()

    bpup_protocol.stop()

    mock_time_changed(
        loop, dt.datetime.now(dt.timezone.utc) + dt.timedelta(seconds=120)
    )
    assert transport.sendto.mock_calls == []


@pytest.mark.asyncio
async def test_protocol_keep_connection_lost_no_error(transport, caplog):
    bpup_subscriptions = BPUPSubscriptions()
    loop = asyncio.get_event_loop()
    bpup_protocol = BPUProtocol(bpup_subscriptions)

    bpup_protocol.connection_made(transport)
    assert transport.sendto.mock_calls == [call(b"\n")]
    transport.sendto.reset_mock()
    assert bpup_subscriptions.alive is False
    bpup_protocol.datagram_received(
        b'{"B":"KNKSADE42149","d":0,"v":"v2.29.2-beta"}\n', MOCK_ADDR
    )
    assert bpup_subscriptions.alive is True

    mock_time_changed(loop, dt.datetime.now(dt.timezone.utc) + dt.timedelta(seconds=60))
    assert transport.sendto.mock_calls == [call(b"\n")]
    transport.sendto.reset_mock()

    mock_protocol_connection_lost(bpup_protocol, None)
    assert "BPUP connection lost" not in caplog.text
    assert bpup_subscriptions.alive is False

    mock_time_changed(
        loop, dt.datetime.now(dt.timezone.utc) + dt.timedelta(seconds=120)
    )
    assert transport.sendto.mock_calls == []


@pytest.mark.asyncio
async def test_protocol_keep_connection_lost_with_error(transport, caplog):
    bpup_subscriptions = BPUPSubscriptions()
    loop = asyncio.get_event_loop()
    bpup_protocol = BPUProtocol(bpup_subscriptions)

    bpup_protocol.connection_made(transport)
    assert transport.sendto.mock_calls == [call(b"\n")]
    transport.sendto.reset_mock()
    assert bpup_subscriptions.alive is False
    bpup_protocol.datagram_received(
        b'{"B":"KNKSADE42149","d":0,"v":"v2.29.2-beta"}\n', MOCK_ADDR
    )

    mock_time_changed(loop, dt.datetime.now(dt.timezone.utc) + dt.timedelta(seconds=60))
    assert transport.sendto.mock_calls == [call(b"\n")]
    transport.sendto.reset_mock()
    assert bpup_subscriptions.alive is True

    mock_protocol_connection_lost(bpup_protocol, OSError())

    assert "BPUP connection lost" in caplog.text
    assert bpup_subscriptions.alive is False

    assert transport.sendto.mock_calls == []
    mock_time_changed(
        loop, dt.datetime.now(dt.timezone.utc) + dt.timedelta(seconds=120)
    )
    assert transport.sendto.mock_calls == []


@pytest.mark.asyncio
async def test_protocol_subscriptions(transport, caplog):
    bpup_subscriptions = BPUPSubscriptions()
    bpup_protocol = BPUProtocol(bpup_subscriptions)
    last_msg = None

    def _on_new_message(msg):
        nonlocal last_msg
        last_msg = msg

    bpup_subscriptions.subscribe("1", _on_new_message)

    bpup_protocol.connection_made(transport)
    # Make sure we can do it again
    bpup_protocol.connection_made(transport)
    bpup_protocol.datagram_received(
        b'{"B":"KNKSADE42149","d":0,"v":"v2.29.2-beta"}\n', MOCK_ADDR
    )

    bpup_protocol.datagram_received(
        b'{"t":"devices/1/state","s":200,"b":{"power":1,"speed":1,"timer":0,"breeze":[0,50,50],"_":"690b6aff"}}\n',
        MOCK_ADDR,
    )

    assert last_msg == {
        "t": "devices/1/state",
        "s": 200,
        "b": {
            "power": 1,
            "speed": 1,
            "timer": 0,
            "breeze": [0, 50, 50],
            "_": "690b6aff",
        },
    }

    bpup_protocol.datagram_received(
        b'{"t":"devices/1/state","s":200,"b":{"power":1,"speed":1,"timer":0,"breeze":[0,50,50],"_":"690b6aff"}}\n',
        MOCK_ADDR,
    )
    # 500 error should not trigger a new message
    assert last_msg == {
        "t": "devices/1/state",
        "s": 200,
        "b": {
            "power": 1,
            "speed": 1,
            "timer": 0,
            "breeze": [0, 50, 50],
            "_": "690b6aff",
        },
    }

    last_msg = {}
    bpup_subscriptions.unsubscribe("1", _on_new_message)
    bpup_protocol.datagram_received(
        b'{"t":"devices/1/state","s":200,"b":{"power":1,"speed":1,"timer":0,"breeze":[0,50,50],"_":"690b6aff"}}\n',
        MOCK_ADDR,
    )
    assert last_msg == {}

    bpup_protocol.datagram_received(
        b'{"B":"KVPRBDGXXXXX","_error_id":633,"_error_msg":"BPUP client timeout"}',
        MOCK_ADDR,
    )
    assert last_msg == {}

    bpup_protocol.datagram_received(
        b"GIGO",
        MOCK_ADDR,
    )
    assert "Failed to process BPUP message" in caplog.text
    assert "GIGO" in caplog.text


@pytest.mark.asyncio
async def test_protocol_errors(transport, caplog):
    bpup_subscriptions = BPUPSubscriptions()
    bpup_protocol = BPUProtocol(bpup_subscriptions)
    bpup_protocol.connection_made(transport)
    bpup_protocol.error_received(OSError())
    assert "BPUP error" in caplog.text


@pytest.mark.asyncio
async def test_start_bpup(transport):
    loop = asyncio.get_event_loop()
    bpup_subscriptions = BPUPSubscriptions()

    async def _mock_create_datagram_endpoint(func, remote_addr=None):
        return transport, func()

    with patch.object(loop, "create_datagram_endpoint", _mock_create_datagram_endpoint):
        stop = await start_bpup("127.0.0.1", bpup_subscriptions)

    stop()



"""Unit tests for Bond API wrapper."""

import pytest
from aiohttp import ClientSession, ClientTimeout
from aioresponses import CallbackResult, aioresponses

from bond_async import Action, Bond, Direction


@pytest.fixture(name="bond")
def bond_fixture():
    """Creates Bond fixture."""
    return Bond("test-host", "test-token")


@pytest.mark.asyncio
async def test_groups_supported(bond: Bond):
    """Tests checking if Bond supports groups."""
    with aioresponses() as response:
        response.get(
            "http://test-host/v2/",
            payload={
                "_": "b7e976b1",
                "__": "00000000",
                "devices": {"_": "0b64ba79"},
                "signal": {"_": "00000000"},
                "groups": {"_": "71ce683b"},
                "sys": {"_": "89858d62"},
                "api": {"_": "d00c7b8f"},
                "bridge": {"_": "df239de9"},
                "token": {"_": "aa72441a"},
            },
        )
        actual = await bond.supports_groups()
        assert actual == True


@pytest.mark.asyncio
async def test_groups_unsupported(bond: Bond):
    """Tests checking if Bond supports groups."""
    with aioresponses() as response:
        response.get(
            "http://test-host/v2/",
            payload={
                "_": "b7e976b1",
                "__": "00000000",
                "devices": {"_": "0b64ba79"},
                "signal": {"_": "00000000"},
                "sys": {"_": "89858d62"},
                "api": {"_": "d00c7b8f"},
                "bridge": {"_": "df239de9"},
                "token": {"_": "aa72441a"},
            },
        )
        actual = await bond.supports_groups()
        assert actual == False


@pytest.mark.asyncio
async def test_groups(bond: Bond):
    """Tests API to get a list of group IDs."""
    with aioresponses() as response:
        response.get(
            "http://test-host/v2/groups",
            payload={
                "_": "some-hash",
                "__": "some-other-hash",
                "group-1": {"_": "some-hash"},
                "group-2": {"_": "some-hash"},
            },
        )
        actual = await bond.groups()
        assert actual == ["group-1", "group-2"]


@pytest.mark.asyncio
async def test_group(bond: Bond):
    """Tests API to get group details."""
    with aioresponses() as response:
        response.get(
            "http://test-host/v2/groups/group-1", payload={"some": "group details"}
        )
        actual = await bond.group("group-1")
        assert actual == {"some": "group details"}


@pytest.mark.asyncio
async def test_group_properties(bond: Bond):
    """Tests API to get group properties."""
    with aioresponses() as response:
        response.get(
            "http://test-host/v2/groups/group-1/properties",
            payload={"some": "group properties"},
        )
        actual = await bond.group_properties("group-1")
        assert actual == {"some": "group properties"}


@pytest.mark.asyncio
async def test_group_state(bond: Bond):
    """Tests API to get group state."""
    with aioresponses() as response:
        response.get(
            "http://test-host/v2/groups/group-1/state",
            payload={"some": "group state"},
        )
        actual = await bond.group_state("group-1")
        assert actual == {"some": "group state"}


@pytest.mark.asyncio
async def test_turn_on(bond: Bond):
    """Tests turn_on action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/TurnOn",
            callback=callback,
        )
        await bond.group_action("test-group-id", Action.turn_on())


@pytest.mark.asyncio
async def test_turn_off(bond: Bond):
    """Tests turn_off action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/TurnOff",
            callback=callback,
        )
        await bond.group_action("test-group-id", Action.turn_off())


@pytest.mark.asyncio
async def test_open(bond: Bond):
    """Tests open action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/Open", callback=callback
        )
        await bond.group_action("test-group-id", Action.open())


@pytest.mark.asyncio
async def test_tilt_open(bond: Bond):
    """Tests tilt open action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/TiltOpen",
            callback=callback,
        )
        await bond.group_action("test-group-id", Action.tilt_open())


@pytest.mark.asyncio
async def test_close(bond: Bond):
    """Tests close action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/Close",
            callback=callback,
        )
        await bond.group_action("test-group-id", Action.close())


@pytest.mark.asyncio
async def test_tilt_close(bond: Bond):
    """Tests tilt close action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/TiltClose",
            callback=callback,
        )
        await bond.group_action("test-group-id", Action.tilt_close())


@pytest.mark.asyncio
async def test_hold(bond: Bond):
    """Tests hold action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/Hold", callback=callback
        )
        await bond.group_action("test-group-id", Action.hold())


@pytest.mark.asyncio
async def test_set_speed(bond: Bond):
    """Tests set_speed action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 2}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/SetSpeed",
            callback=callback,
        )
        await bond.group_action("test-group-id", Action.set_speed(2))


@pytest.mark.asyncio
async def test_set_speed_belief(bond: Bond):
    """Tests set_speed_belief action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"speed": 2}
            return CallbackResult()

        response.patch(
            "http://test-host/v2/groups/test-group-id/state", callback=callback
        )
        await bond.group_action("test-group-id", Action.set_speed_belief(2))


@pytest.mark.asyncio
async def test_set_brightness_belief(bond: Bond):
    """Tests set_brightness_belief action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"brightness": 2}
            return CallbackResult()

        response.patch(
            "http://test-host/v2/groups/test-group-id/state", callback=callback
        )
        await bond.group_action("test-group-id", Action.set_brightness_belief(2))


@pytest.mark.asyncio
async def test_set_power_state_belief(bond: Bond):
    """Tests set_power_state_belief action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"power": 1}
            return CallbackResult()

        response.patch(
            "http://test-host/v2/groups/test-group-id/state", callback=callback
        )
        await bond.group_action("test-group-id", Action.set_power_state_belief(True))


@pytest.mark.asyncio
async def test_set_light_state_belief(bond: Bond):
    """Tests set_light_state_belief action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"light": 0}
            return CallbackResult()

        response.patch(
            "http://test-host/v2/groups/test-group-id/state", callback=callback
        )
        await bond.group_action("test-group-id", Action.set_light_state_belief(False))


@pytest.mark.asyncio
async def test_turn_light_on(bond: Bond):
    """Tests turn_light_on action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/TurnLightOn",
            callback=callback,
        )
        await bond.group_action("test-group-id", Action.turn_light_on())


@pytest.mark.asyncio
async def test_turn_light_off(bond: Bond):
    """Tests turn_light_off action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/TurnLightOff",
            callback=callback,
        )
        await bond.group_action("test-group-id", Action.turn_light_off())


@pytest.mark.asyncio
async def test_set_brightness(bond: Bond):
    """Tests set_brightness action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 50}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/SetBrightness",
            callback=callback,
        )
        await bond.group_action("test-group-id", Action.set_brightness(50))


@pytest.mark.asyncio
async def test_set_color_temperature(bond: Bond):
    """Tests set_color_temperature action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 3000}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/SetColorTemp",
            callback=callback,
        )
        await bond.group_action("test-group-id", Action.set_color_temperature(3000))


@pytest.mark.asyncio
async def test_increase_color_temperature(bond: Bond):
    """Tests increase_color_temperature action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 100}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/IncreaseColorTemp",
            callback=callback,
        )
        await bond.group_action("test-group-id", Action.increase_color_temperature(100))


@pytest.mark.asyncio
async def test_decrease_color_temperature(bond: Bond):
    """Tests decrease_color_temperature action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 100}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/DecreaseColorTemp",
            callback=callback,
        )
        await bond.group_action("test-group-id", Action.decrease_color_temperature(100))


@pytest.mark.asyncio
async def test_set_direction_forward(bond: Bond):
    """Tests set_direction action delegates to API with correct value for forward."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 1}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/SetDirection",
            callback=callback,
        )
        await bond.group_action(
            "test-group-id", Action.set_direction(Direction.FORWARD)
        )


@pytest.mark.asyncio
async def test_set_direction_reverse(bond: Bond):
    """Tests set_direction action delegates to API with correct value for reverse."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": -1}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/SetDirection",
            callback=callback,
        )
        await bond.group_action(
            "test-group-id", Action.set_direction(Direction.REVERSE)
        )


@pytest.mark.asyncio
async def test_set_flame(bond: Bond):
    """Tests set_flame action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 50}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/SetFlame",
            callback=callback,
        )
        await bond.group_action("test-group-id", Action.set_flame(50))


@pytest.mark.asyncio
async def test_set_position(bond: Bond):
    """Tests set_position action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 50}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/SetPosition",
            callback=callback,
        )
        await bond.group_action("test-group-id", Action.set_position(50))


@pytest.mark.asyncio
async def test_increase_position(bond: Bond):
    """Tests increase_position action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 50}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/IncreasePosition",
            callback=callback,
        )
        await bond.group_action("test-group-id", Action.increase_position(50))


@pytest.mark.asyncio
async def test_decrease_position(bond: Bond):
    """Tests decrease_position action delegates to API."""
    with aioresponses() as response:

        def callback(_url, **kwargs):
            assert kwargs.get("json") == {"argument": 50}
            return CallbackResult()

        response.put(
            "http://test-host/v2/groups/test-group-id/actions/DecreasePosition",
            callback=callback,
        )
        await bond.group_action("test-group-id", Action.decrease_position(50))



"""Setup script for bond-async."""

import pathlib

from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name="bond_async",
    version="0.2.1",
    packages=find_packages(exclude=['tests']),

    author="Olibra LLC",
    author_email="mobiledev@olibra.io",
    description="Asynchronous Python wrapper library over Bond Local API",
    long_description=README,
    long_description_content_type="text/markdown",
    keywords="bond local api async",

    install_requires=[
        "orjson>=3.8.7",
        "aiohttp>=3.6.1"
    ],

    url="https://github.com/bondhome/bond-async",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Topic :: Home Automation"
    ]
)



"""Asynchronous Python wrapper library over Bond Local API."""

from .action import Action, Direction
from .bond import Bond
from .bond_type import BondType
from .bpup import BPUPSubscriptions, start_bpup
from .device_type import DeviceType
from .bond_type import BondType
from .requestor_uuid import RequestorUUID

__all__ = [
    "Bond",
    "BPUPSubscriptions",
    "start_bpup",
    "Action",
    "Direction",
    "DeviceType",
    "BondType",
    "RequestorUUID",
]



"""Unit tests for DeviceType."""

from bond_async.device_type import DeviceType


def test_compare_device_types():
    """Tests that constants are configured and compared correctly."""
    assert DeviceType.CEILING_FAN == "CF"
    assert DeviceType.is_fan("CF")

    assert DeviceType.MOTORIZED_SHADES == "MS"
    assert DeviceType.is_shades("MS")

    assert DeviceType.FIREPLACE == "FP"
    assert DeviceType.is_fireplace("FP")

    assert DeviceType.GENERIC_DEVICE == "GX"
    assert DeviceType.is_generic("GX")

    assert DeviceType.AIR_CONDITIONER == "AC"
    assert DeviceType.is_air_conditioner("AC")

    assert DeviceType.GARAGE_DOOR == "GD"
    assert DeviceType.is_garage_door("GD")

    assert DeviceType.BIDET == "BD"
    assert DeviceType.is_bidet("BD")

    assert DeviceType.LIGHT == "LT"
    assert DeviceType.is_light("LT")
